' ----------------------------------------------------------------------------------
' Microsoft Developer & Platform Evangelism
' 
' Copyright (c) Microsoft Corporation. All rights reserved.
' 
' THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
' EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
' OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
' ----------------------------------------------------------------------------------
' The example companies, organizations, products, domain names,
' e-mail addresses, logos, people, places, and events depicted
' herein are fictitious.  No association with any real company,
' organization, product, domain name, email address, logo, person,
' places, or events is intended or should be inferred.
' ----------------------------------------------------------------------------------

Option Compare Binary
Option Infer On
Option Strict On
Option Explicit On

Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.ComponentModel.DataAnnotations
Imports System.Linq
Imports System.ServiceModel.DomainServices.Hosting
Imports System.ServiceModel.DomainServices.Server
Imports System.Configuration
Imports System.Web.Configuration


'TODO: Create methods containing your application logic.
'TODO: add the EnableClientAccess custom attribute to this class to expose this DomainService to clients.


Public Class ProductsService
    Inherits DomainService

    Private _sourceFile As XElement
    Private _filePath As String

    Public Overrides Sub Initialize(ByVal context As System.ServiceModel.DomainServices.Server.DomainServiceContext)
        MyBase.Initialize(context)

        'get the file path from the config

        If WebConfigurationManager.ConnectionStrings("XMLProducts.ProductsFilePath") Is Nothing Then
            Throw New Exception("The filepath must be provided in the web.config in the connection strings section under XMLProducts.ProductsFilePath.")
        Else
            _filePath = WebConfigurationManager.ConnectionStrings("XMLProducts.ProductsFilePath").ConnectionString
            If Not System.IO.File.Exists(_filePath) Then
                Dim x As XElement = New XElement("DataRoot")
                x.Save(_filePath)
            End If

            _sourceFile = XElement.Load(_filePath)

            'Verify file
            If _sourceFile.Name <> "DataRoot" Then
                Throw New Exception("Corrupt file.")
            End If

            'create a product categories node if one does not exist
            If _sourceFile.<ProductCategories>.Count = 0 Then
                _sourceFile.Add(New XElement("ProductCategories"))
            End If

            'create a categories node if one does not exist
            If _sourceFile.<Products>.Count = 0 Then
                _sourceFile.Add(New XElement("Products"))
            End If
        End If
    End Sub

    Public Overrides Function Submit(ByVal changeSet As System.ServiceModel.DomainServices.Server.ChangeSet) As Boolean
        'reorder changeset to ensure that categories are saved first
        Dim c As New ChangeSet(changeSet.ChangeSetEntries.OrderBy(Function(entry) entry.Entity, New ProductsEntitiesComparer))

        Dim baseResult As Boolean = MyBase.Submit(c)
        If baseResult Then
            _sourceFile.Save(_filePath)
            Return True
        Else
            Return False
        End If
    End Function


#Region "Queries"
    <Query(IsDefault:=True)> _
    Public Function GetProducts() As IEnumerable(Of Product)
        Dim products As New List(Of Product)
        For Each pElem As XElement In _sourceFile...<Product>
            Dim p As New Product
            p.ProductID = New Guid(pElem.<ProductID>.Value)
            p.ProductName = pElem.<ProductName>.Value
            p.CategoryID = New Guid(pElem.<CategoryID>.Value)
            p.QuantityPerUnit = pElem.<QuantityPerUnit>.Value
            p.UnitPrice = CDec(pElem.<UnitPrice>.Value)
            p.UnitsInStock = CInt(pElem.<UnitsInStock>.Value)
            p.ReorderLevel = CInt(pElem.<ReorderLevel>.Value)
            p.Discontinued = CBool(pElem.<Discontinued>.Value)
            products.Add(p)

        Next
        Return products
    End Function

    <Query(IsDefault:=True)> _
    Public Function GetProductCategories() As IEnumerable(Of ProductCategory)
        Dim categories As New List(Of ProductCategory)

        For Each catElem As XElement In _sourceFile...<ProductCategory>
            Dim pc As New ProductCategory()
            pc.CategoryID = New Guid(catElem.<CategoryID>.Value)
            pc.CategoryName = catElem.<CategoryName>.Value
            pc.Description = catElem.<Description>.Value
            categories.Add(pc)
        Next
        Return categories
    End Function
#End Region

#Region "Product Update/Delete/Insert Methods"
    Public Sub InsertProduct(ByVal p As Product)
        Try
            'generate a new guid for the product
            p.ProductID = Guid.NewGuid

            Dim productElem As XElement = _
                <Product>
                    <ProductID DataType="Guid"><%= p.ProductID %></ProductID>
                    <ProductName DataType="String"><%= p.ProductName %></ProductName>
                    <CategoryID DataType="Guid"><%= p.CategoryID %></CategoryID>
                    <QuantityPerUnit DataType="String"><%= p.QuantityPerUnit %></QuantityPerUnit>
                    <UnitPrice DataType="Decimal"><%= p.UnitPrice %></UnitPrice>
                    <UnitsInStock DataType="Int32"><%= p.UnitsInStock %></UnitsInStock>
                    <ReorderLevel DataType="Int32"><%= p.ReorderLevel %></ReorderLevel>
                    <Discontinued DataType="Boolean"><%= p.Discontinued %></Discontinued>
                </Product>

            _sourceFile.<Products>.First.Add(productElem)
        Catch ex As Exception
            Throw New Exception("Error inserting Product " + p.ProductName, ex)
        End Try
    End Sub

    Public Sub UpdateProduct(ByVal p As Product)
        Try
            'get the existing item from the db
            Dim storeProductElem As XElement =
                (From c As XElement In _sourceFile...<Product> Where c.<ProductID>.Value = p.ProductID.ToString).FirstOrDefault

            If storeProductElem Is Nothing Then
                'product does not exist.  Need to indicate that the item has already been deleted.
                'Need to find the item in the changeset
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is p _
                    Select e).First
                entry.IsDeleteConflict = True
            Else
                'comparison of elements to generate concurrency errors
                Dim storeProduct As New Product()
                storeProduct.ProductID = New Guid(storeProductElem.<ProductID>.Value)
                storeProduct.ProductName = storeProductElem.<ProductName>.Value
                storeProduct.CategoryID = New Guid(storeProductElem.<CategoryID>.Value)
                storeProduct.QuantityPerUnit = storeProductElem.<QuantityPerUnit>.Value
                storeProduct.UnitPrice = CDec(storeProductElem.<UnitPrice>.Value)
                storeProduct.UnitsInStock = CInt(storeProductElem.<UnitsInStock>.Value)
                storeProduct.ReorderLevel = CInt(storeProductElem.<ReorderLevel>.Value)
                storeProduct.Discontinued = CBool(storeProductElem.<Discontinued>.Value)

                'Need to find the item in the changeset to compare original values
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is p _
                    Select e).First

                'list to track conflicting fields
                Dim conflictMembers As New List(Of String)

                If storeProduct.ProductName <> DirectCast(entry.OriginalEntity, Product).ProductName Then
                    conflictMembers.Add("ProductName")
                End If
                If storeProduct.CategoryID <> DirectCast(entry.OriginalEntity, Product).CategoryID Then
                    conflictMembers.Add("CategoryID")
                End If
                If storeProduct.QuantityPerUnit <> DirectCast(entry.OriginalEntity, Product).QuantityPerUnit Then
                    conflictMembers.Add("QuantityPerUnit")
                End If
                If storeProduct.UnitPrice <> DirectCast(entry.OriginalEntity, Product).UnitPrice Then
                    conflictMembers.Add("UnitPrice")
                End If
                If storeProduct.UnitsInStock <> DirectCast(entry.OriginalEntity, Product).UnitsInStock Then
                    conflictMembers.Add("UnitsInStock")
                End If
                If storeProduct.ReorderLevel <> DirectCast(entry.OriginalEntity, Product).ReorderLevel Then
                    conflictMembers.Add("ReorderLevel")
                End If
                If storeProduct.Discontinued <> DirectCast(entry.OriginalEntity, Product).Discontinued Then
                    conflictMembers.Add("Discontinued")
                End If

                'set the conflict members
                entry.ConflictMembers = conflictMembers
                entry.StoreEntity = storeProduct

                If conflictMembers.Count < 1 Then
                    'update the xml db
                    Dim productElem As XElement = _
                        <Product>
                            <ProductID DataType="Guid"><%= p.ProductID %></ProductID>
                            <ProductName DataType="String"><%= p.ProductName %></ProductName>
                            <CategoryID DataType="Guid"><%= p.CategoryID %></CategoryID>
                            <QuantityPerUnit DataType="String"><%= p.QuantityPerUnit %></QuantityPerUnit>
                            <UnitPrice DataType="Decimal"><%= p.UnitPrice %></UnitPrice>
                            <UnitsInStock DataType="Int32"><%= p.UnitsInStock %></UnitsInStock>
                            <ReorderLevel DataType="Int32"><%= p.ReorderLevel %></ReorderLevel>
                            <Discontinued DataType="Boolean"><%= p.Discontinued %></Discontinued>
                        </Product>
                    storeProductElem.ReplaceWith(productElem)
                End If
            End If
        Catch ex As Exception
            Throw New Exception("Error updating Product " + p.ProductID.ToString + ":" + p.ProductName, ex)
        End Try
    End Sub

    Public Sub DeleteProduct(ByVal p As Product)
        Try
            'get the existing item from the db
            Dim storeProductElem As XElement =
                (From c As XElement In _sourceFile...<Product> Where c.<ProductID>.Value = p.ProductID.ToString).FirstOrDefault

            If storeProductElem Is Nothing Then
                'product category does not exist.  Need to indicate that the item has already been deleted.
                'Need to find the item in the changeset
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is p _
                    Select e).First
                entry.IsDeleteConflict = True
            Else
                'remove it
                storeProductElem.Remove()
            End If
        Catch ex As Exception
            Throw New Exception("Error deleting Product " + p.ProductID.ToString + ":" + p.ProductName, ex)
        End Try
    End Sub
#End Region

#Region "Category Update/Delete/Insert Methods"
    Public Sub InsertProductCategory(ByVal pc As ProductCategory)
        Try
            'generate a new Guid for the category
            pc.CategoryID = Guid.NewGuid

            Dim categoryElem As XElement = _
                <ProductCategory>
                    <CategoryID DataType="Guid"><%= pc.CategoryID %></CategoryID>
                    <CategoryName DataType="String"><%= pc.CategoryName %></CategoryName>
                    <Description DataType="String"><%= pc.Description %></Description>
                </ProductCategory>


            'update the category ID on any related product entities
            For Each e As ChangeSetEntry In ChangeSet.ChangeSetEntries
                If TypeOf (e.Entity) Is Product AndAlso DirectCast(e.Entity, Product).Category Is pc Then
                    DirectCast(e.Entity, Product).CategoryID = pc.CategoryID
                End If
            Next

            'update the xml doc
            _sourceFile.<ProductCategories>.First.Add(categoryElem)
        Catch ex As Exception
            Throw New Exception("Error inserting ProductCategory " + pc.CategoryName, ex)
        End Try


    End Sub

    Public Sub UpdateProductCategory(ByVal pc As ProductCategory)
        Try
            'get the existing item from the db
            Dim storeCategoryElem As XElement =
                (From c As XElement In _sourceFile...<ProductCategory> Where c.<CategoryID>.Value = pc.CategoryID.ToString).FirstOrDefault

            If storeCategoryElem Is Nothing Then
                'product category does not exist.  Need to indicate that the item has already been deleted.
                'Need to find the item in the changeset
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is pc _
                    Select e).First
                entry.IsDeleteConflict = True
            Else
                'comparison of elements to generate concurrency errors
                Dim storeCategory As New ProductCategory()
                storeCategory.CategoryID = New Guid(storeCategoryElem.<CategoryID>.Value)
                storeCategory.CategoryName = storeCategoryElem.<CategoryName>.Value
                storeCategory.Description = storeCategoryElem.<Description>.Value

                'Need to find the item in the changeset to compare original values
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is pc _
                    Select e).First

                'list to track conflicting fields
                Dim conflictMembers As New List(Of String)

                If storeCategory.CategoryName <> DirectCast(entry.OriginalEntity, ProductCategory).CategoryName Then
                    conflictMembers.Add("CategoryName")
                End If
                If storeCategory.Description <> DirectCast(entry.OriginalEntity, ProductCategory).Description Then
                    conflictMembers.Add("Description")
                End If

                'set the conflict members
                entry.ConflictMembers = conflictMembers
                entry.StoreEntity = storeCategory

                If conflictMembers.Count < 1 Then
                    'update the database file
                    Dim categoryElem As XElement = _
                    <ProductCategory>
                        <CategoryID DataType="Guid"><%= pc.CategoryID %></CategoryID>
                        <CategoryName DataType="String"><%= pc.CategoryName %></CategoryName>
                        <Description DataType="String"><%= pc.Description %></Description>
                    </ProductCategory>
                    storeCategoryElem.ReplaceWith(categoryElem)
                End If
            End If
        Catch ex As Exception
            Throw New Exception("Error updating ProductCategory " + pc.CategoryID.ToString + ":" + pc.CategoryName, ex)
        End Try

    End Sub

    Public Sub DeleteProductCategory(ByVal pc As ProductCategory)
        Try
            'get the existing item from the db
            Dim storeCategoryElem As XElement =
                (From c As XElement In _sourceFile...<ProductCategory> Where c.<CategoryID>.Value = pc.CategoryID.ToString).FirstOrDefault

            If storeCategoryElem Is Nothing Then
                'product category does not exist.  Need to indicate that the item has already been deleted.
                'Need to find the item in the changeset
                Dim entry As ChangeSetEntry = _
                    (From e As ChangeSetEntry In ChangeSet.ChangeSetEntries _
                    Where e.Entity Is pc _
                    Select e).First
                entry.IsDeleteConflict = True
            Else
                '<TODO> check if there are any existing products with this category. Don't allow delete in this case
                'need to get related products from data source and ensure that they are not already in change list for deletion

                'Dim relatedProductsCount As Integer = _
                '    (From c As XElement In _db...<Product> Where c.<CategoryID>.Value = pc.CategoryID.ToString).Count

                'If relatedProductsCount > 0 Then
                '    Throw New Exception("There are existing products for this category.")
                'End If

                'remove it
                storeCategoryElem.Remove()
            End If
        Catch ex As Exception
            Throw New Exception("Error deleting ProductCategory " + pc.CategoryID.ToString + ":" + pc.CategoryName, ex)
        End Try

    End Sub
#End Region


    Public Class ProductsEntitiesComparer
        Inherits Comparer(Of Object)

        Public Overrides Function Compare(ByVal x As Object, ByVal y As Object) As Integer
            If TypeOf (x) Is Product And TypeOf (y) Is ProductCategory Then
                Return 1
            ElseIf TypeOf (x) Is ProductCategory And TypeOf (y) Is Product Then
                Return -1
            Else
                Return 0
            End If
        End Function
    End Class
End Class